stocks_list_renewable = ['FSLR', 'DQ', 'CWEN','BYD', 'ALB', 'LTHM', 'LAC', 'TSLA', 'RIVN'] # solar, battery, ecar
stock_list_tech = ['AAPL', 'MSFT', 'NVDA', 'AVGO', 'AMD', 'ADBE', 'AMAT', 'GOOGL'] # tech
stocks_list_rareearth = ['NYSE:MP', 'TSX:EFR', 'TSX:NB', 'TSX:ARA', 'TSXV:UCU', 'ASX:LYC', 'ASX:ILU', 'ASX:ARU'] # rare earth
stocks_list_all = stocks_list_renewable + stock_list_tech + stocks_list_rareearth